"""Base Rule module"""

from typing import (
    Set,
    Dict,
    Any,
    Callable,
    Optional,
)

from abc import (
    ABC,
    abstractmethod,
)

from .context import Context
from .dr_property import dr_property


class Rule(ABC):
    """Base Rule class for dictrule module"""

    def __init__(self) -> None:
        self._dr_props = set(dr_property.properties(self))
        self._dr_non_optional_props = set(dr_property.properties(self, optional=False))

    @property
    def dr_props(self) -> Set[Callable]:
        """All `dr_property` properties

        Returns:
            Set[Callable]: Set of `dr_property` functions
        """

        return self._dr_props

    @property
    def dr_non_optional_props(self) -> Set[Callable]:
        """Non optional `dr_property` properties

        Returns:
            Set[Callable]: Set of non-optional `dr_property` functions
        """

        return self._dr_non_optional_props

    @abstractmethod
    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context]], "Rule"],
        context: Optional[Context] = None,
    ) -> str:
        """Abstract parse method for Rule

        Args:
            rule_dict (Dict[str, Any]): Dictionay of rules to generate
            rule_callback (Callable[[Optional[Context]], Rule]): rule callback
                for rules not handled by the current rule
            context (Optional[Context], optional): Context for the rule. Defaults to None.

        Returns:
            str: Generated text
        """

        return ""



"""For-in rule module"""

from typing import (
    Dict,
    List,
    Iterable,
    Any,
    Callable,
    Optional,
)

from .eval_rule import EvalRule
from .block_rule import BlockRule
from ..rule import Rule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    NoneValueException,
    InvalidTypeException,
)


class ForInRule(Rule):
    """This rule executes generatable rules in a `for-in-block` loop
    with a provided iterable variable.

    Requires providing a `Context` that includes the necessary information
    for the rules in the `block`.

    The `in` keyword must be represented by a `EvalRule` variable,
    and the value of this variable must be a `Interable` type.

    Examples:
    ---------
    >>> context = build_saved_lines_context([
    ...     Line(content="This is the line_1 content"),
    ...     Line(content="This is the line_2 content"),
    ... ])
    >>> dictrule.Generator({
    ...     "for": "line"
    ...     "in": "saved_lines"
    ...     "block: [
    ...         "line.index",
    ...         "line.content"
    ...     ]
    ... }).generate()
    1
    This is the line_1 content
    2
    This is the line_2 content
    """

    @dr_property()
    def _for(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `for` attribute."""

    @dr_property()
    def _in(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `in` attribute."""

    @dr_property()
    def _block(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `block` attribute."""

    class ForInEval(EvalRule.Evaluable):
        """Rule for executing generable rules in a `for-in-eval` loop.

        Examples:
        ---------
        command_list = [
            "This Is The Title",
            "and this is the description",
            "are_created_at",
            "2024-01-01",
        ]

        >>> rule_dict = {
        ...     "for": "command",
        ...     "in": "command_list",
        ...     "eval": "command",
        ... }
        This Is The Title
        and this is the description
        are_created_at
        2024-01-01
        """

        @property
        def name(self) -> str:
            """Getter for `name` property"""

            return self._var_name

        @property
        def prefix_matching(self) -> bool:
            """Getter for `prefix_matching` property"""

            return True

        def __init__(
            self,
            var_name: str,
            var: Any,
            extra_properties: Dict[str, Any],
        ):
            """Initializes the ForInEval object.

            Args:
                var_name (str): The name of the variable in the rule dictionary.
                var (Any): The name of the executable variable.
                extra_properties (Dict[str, Any]): A map of variable names and values.
            """

            self._var_name = str(var_name)
            self._var = var
            self._extra_properties = dict(extra_properties)

        def run(
            self,
            cmd: str,
        ) -> str:
            """Executes the evaluable rule.

            This method runs the evaluable rule based on the provided command.

            Args:
                cmd (str): The command to be executed.

            Returns:
                str: The result of the execution.
            """

            var_prefix = self._var_name + "."
            if not cmd.startswith(var_prefix):
                return self._var

            properties = cmd.split(".")
            property_value = ""
            extra_prop = ".".join(properties[1:])
            if extra_prop in self._extra_properties:
                property_value = self._extra_properties[extra_prop]
            else:
                local_var = self._var
                for prop in properties[1:]:
                    callable_var = getattr(local_var, prop)
                    if callable(callable_var):
                        local_var = callable_var(local_var)
                    else:
                        local_var = callable_var

                property_value = local_var

            return property_value

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parses the rule dictionary for `ForInRule`.

        Args:
            rule_dict (Dict[str, Any]): Dictionary of rules containing `ForInRule`.
            rule_callback (Callable[[Optional[Context], Any], str]): Fallback for other rules.
            context (Optional[Context], optional): `ForInRule` context. Defaults to None.

        Returns:
            str: Parsed value for `rule_dict`.
        """

        if context is None:
            raise NoneValueException("param `context` must not be None")

        _, for_var = self._for(rule_dict)
        _, in_var = self._in(rule_dict)
        _, block = self._block(rule_dict)

        if not isinstance(for_var, str):
            raise InvalidTypeException(f"`for:` {for_var} must be a str")

        if not isinstance(in_var, str):
            raise InvalidTypeException(f"`for:in:` {in_var} must be a str")

        if not isinstance(block, List):
            raise InvalidTypeException(f"`for:block:` {block} must be a list")

        eval_in = EvalRule().internal_parse(
            rule_dict={
                "eval": in_var,
            },
            context=context,
            rule_callback=rule_callback,
        )

        if not isinstance(eval_in, Iterable):
            raise InvalidTypeException("`in` must return an Iterable value")

        blocks: List[str] = []
        case_map = dict(context.case_map)
        eval_context_case: EvalRule.ContextCase = case_map.get(EvalRule.CONTEXT_NAME)
        is_eval_context_case = isinstance(eval_context_case, EvalRule.ContextCase)
        case_map.pop(EvalRule.CONTEXT_NAME)
        without_eval_cases = list(case_map.values())
        for index, var in enumerate(eval_in):
            block_context = Context(
                without_eval_cases
                + [
                    EvalRule.ContextCase(
                        evaluators=(
                            eval_context_case.evaluator_list
                            if is_eval_context_case
                            else []
                        )
                        + [
                            ForInRule.ForInEval(
                                var_name=for_var,
                                var=var,
                                extra_properties={
                                    "index": index,
                                },
                            )
                        ],
                        fallback=(
                            eval_context_case.fallback if is_eval_context_case else None
                        ),
                    ),
                ],
            )

            block_parsed = BlockRule().parse(
                rule_dict={
                    "block": block,
                },
                rule_callback=rule_callback,
                context=block_context,
            )

            blocks.append(block_parsed)

        return "\n".join(blocks)



"""Indent rule module"""

from typing import (
    Dict,
    Any,
    Callable,
    Optional,
)

from ..rule import Rule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    NoneValueException,
    InvalidValueException,
    InvalidTypeException,
)


class IndentRule(Rule):
    """This rule indents generated text from provided rules by a specified number of indent spaces.

    Defines the number of spaces for text indentation by providing `IndentRule.ContextCase`
    within the `Context`.

    Examples:
    ---------
    >>> context = build_number_spaces_context(4)
    >>> dictrule.Generator({
    ...     "class Menu:"
    ...     "indent_1": "def content(self):"
    ...     "indent_2": "return 'Empty'"
    ... }).generate(context)
    class Menu:
        def content(self):
            return 'Empty'
    """

    DEFAULT_SPACES = 2
    CONTEXT_NAME = "indent"

    class ContextCase(Context.Case):
        """`Context.Case` for `IndentRule`."""

        @property
        def name(self) -> str:
            """Getter for `name` property"""

            return "indent"

        def __init__(
            self,
            num_spaces: int,
        ):
            """Initializes the context case.

            Args:
                num_spaces (int): The number of spaces for each level of indentation.
            """

            self._num_spaces = num_spaces

        @property
        def num_spaces(self) -> int:
            """Getter for `num_spaces` property"""

            return self._num_spaces

    @dr_property(prefix_matching=True)
    def _indent(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `indent` attribute."""

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parses the indent rule and applies indentation to the provided text.

        Args:
            rule_dict (Dict[str, Any]): The dictionary containing the indent rule.
            rule_callback (Callable[[Optional[Context], Any], str]): A callback function
                for processing rules.
            context (Optional[Context], optional): The context for the rule. Defaults to None.

        Returns:
            str: The indented text.
        """

        if context is None:
            raise NoneValueException("param `context` must not be None")

        context_case: IndentRule.ContextCase = context.get(IndentRule.CONTEXT_NAME)

        indent_spaces = IndentRule.DEFAULT_SPACES
        if context_case:
            if not isinstance(context_case, IndentRule.ContextCase):
                raise InvalidTypeException(
                    f"Invalid {type(context_case)} type for {IndentRule.CONTEXT_NAME} in context"
                )

            indent_spaces = context_case.num_spaces

        indent, value = self._indent(rule_dict)
        if not indent.startswith("indent_"):
            raise InvalidValueException(f"Invalid indent {indent}")

        if not indent_spaces or indent_spaces <= 0:
            indent_spaces = IndentRule.DEFAULT_SPACES

        if not isinstance(indent_spaces, int) and not (
            isinstance(indent_spaces, str) and str.isdigit(indent_spaces)
        ):
            raise InvalidTypeException("`indent_spaces` must be a digit")

        indent_spaces: int = int(indent_spaces)

        if not indent_spaces:
            print("`indent_spaces` is not found in `context`, using default spaces")
            indent_spaces = 4

        if not isinstance(value, str):
            value = rule_callback(context, value)

        indent_prefix = indent_spaces
        if indent != "indent":
            indent_count = indent[len("indent_") :]
            if not indent_count.isdigit():
                raise InvalidValueException("`indent_` suffix must be a digit str")

            indent_prefix = indent_spaces * int(indent_count) * " "

        value = indent_prefix + value.replace("\n", f"\n{indent_prefix}")
        return value



"""Exception module"""


class InvalidValueException(Exception):
    """Invalid value exception"""


class NoneValueException(Exception):
    """None value exception"""


class InvalidTypeException(Exception):
    """Invalid type exception"""



"""dr_property decorator module"""

import functools
from typing import (
    List,
    Dict,
    Any,
    Callable,
    Optional,
)

from .exceptions import (
    InvalidTypeException,
)


class dr_property:
    """dictrule decorator that defines property for rules.
    Supports prefix matching and optional properties

    Properties decorated with `dr_property` must have the `dr_property._prefix` prefix.

    Usage:
        - Defines:
            @dr_property(prefix_matching=True, optional=True)
            def _eval(self, props: Dict[str, Any]) -> Any:
                pass
        - Parses:
            key, value = self._eval(rule_dict)

    """

    _prefix = "_"
    _flag_key = "_is_dr_property"
    _name_key = "_dr_name"
    _optional_key = "_dr_optional"

    def __init__(
        self,
        optional: bool = False,
        prefix_matching: bool = False,
    ):
        """Initialization method of `dr_property`

        Args:
            optional (bool, optional): Indicates if the property is optional
                and used for detecting rules. Defaults to False.
            prefix_matching (bool, optional): Indicates if the property matches the rule prefix.
                Defaults to False.
        """

        self._optional = optional
        self._prefix_matching = prefix_matching

    def __call__(
        self,
        *args,
        **kwds,
    ) -> Optional[Any]:
        func = args[0]
        key = func.__name__[len(dr_property._prefix) :]
        setattr(func, dr_property._flag_key, True)
        setattr(func, dr_property._name_key, key)
        setattr(func, dr_property._optional_key, self._optional)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func_value = func(*args, **kwargs)
            if func_value:
                return func_value

            props: Dict[str, Any] = args[1]
            if not isinstance(props, Dict):
                raise InvalidTypeException(f"First param `{props}` must be a dict")

            if self._prefix_matching:
                for prop, value in props.items():
                    if prop.startswith(key):
                        return (prop, value)

            return (key, props.get(key))

        return wrapper

    @classmethod
    def properties(
        cls,
        instance: Any,
        optional: bool = True,
    ) -> List[Callable]:
        """Fetches all `dr_property` of `instance`.

        Args:
            instance (Any): Instance using `dr_property`.
            optional (bool, optional): True if fetching optional `dr_property`. Defaults to True.

        Returns:
            List[Callable]: List of `dr_property` functions
        """
        properties: List[Callable] = []
        for name in dir(instance):
            attr = getattr(instance, name, None)
            is_tg = getattr(attr, cls._flag_key, None)
            if not is_tg:
                continue

            if not optional:
                is_optional = getattr(attr, cls._optional_key, None)
                if is_optional:
                    continue

            properties.append(attr)
        return properties

    @classmethod
    def str_properties(
        cls,
        instance: Any,
        optional: bool = True,
    ) -> List[str]:
        """Fetches all `dr_property` of `instance` as list of strings.

        Args:
            instance (Any): Instance using `dr_property`
            optional (bool, optional): True if fetching optional `dr_property`. Defaults to True.

        Returns:
            List[Callable]: List of `dr_property` strings.
        """
        return [
            getattr(prop, dr_property._name_key)
            for prop in cls.properties(
                instance=instance,
                optional=optional,
            )
        ]



"""Format rule module"""

from typing import (
    Dict,
    Any,
    Callable,
    Optional,
)

from enum import Enum
from ..rule import Rule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    NoneValueException,
    InvalidValueException,
    InvalidTypeException,
)


class FormatRule(Rule):
    """This rule formats text according to specified rules.

    Examples:
    ---------
    >>> dictrule.Generator({
    ...     "format_lowercase": "UPPERCASE TITLE"
    ...     "format_uppercase": "lowercase content"
    ...     "format_upper_head": "camelVariable"
    ... }).generated()
    uppercase title
    LOWERCASE CONTENT
    CamelVariable
    """

    @dr_property(prefix_matching=True)
    def _format(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `format` attribute."""

    class Type(Enum):
        """Enum representing different formatting types for `FormatRule`."""

        LOWERCASE = "lowercase"
        UPPERCASE = "uppercase"
        UPPER_HEAD = "upper_head"

        @staticmethod
        def from_str(string: str) -> Optional["FormatRule.Type"]:
            """Creates a `FormatRule.Type` enum from a string.

            Args:
                string (str): The string value representing the formatting type.

            Returns:
                Optional[FormatRule.Type]: The corresponding enum value.
            """

            try:
                return FormatRule.Type(string)
            except ValueError:
                pass

            return None

        def format(
            self,
            text: str,
        ) -> str:
            """Formats the provided text based on the formatting type.

            Args:
                text (str): The text to be formatted.

            Returns:
                str: The formatted text.
            """

            if self == FormatRule.Type.LOWERCASE:
                return text.lower()

            if self == FormatRule.Type.UPPERCASE:
                return text.upper()

            if self == FormatRule.Type.UPPER_HEAD:
                return (text[0].upper() + text[1:]) if len(text) > 0 else ""

            return text

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parses the format rule and applies formatting to the provided text.

        Args:
            rule_dict (Dict[str, Any]): The dictionary containing the format rule.
            rule_callback (Callable[[Optional[Context], Any], str]): A callback function
                for processing rules.
            context (Optional[Context], optional): The context for the rule. Defaults to None.

        Returns:
            str: The formatted text.
        """

        format_name, format_rule = self._format(rule_dict)
        if not format_name.startswith("format_"):
            raise InvalidValueException(f"Invalid format {format_name}")

        format_type = FormatRule.Type.from_str(format_name[len("format_") :])
        if format_type is None:
            raise NoneValueException(
                "`format:` type {format_type} in invalid with format {format_name}"
            )

        format_text = rule_callback(
            context,
            format_rule,
        )

        if not isinstance(format_text, str):
            raise InvalidTypeException(
                f"`format:` text {format_text} for rule {format_rule} must be a str"
            )

        return format_type.format(format_text)



r"""
       ___      __             __   
  ____/ (_)____/ /________  __/ /__ 
 / __  / / ___/ __/ ___/ / / / / _ \
/ /_/ / / /__/ /_/ /  / /_/ / /  __/
\__,_/_/\___/\__/_/   \__,_/_/\___/ 

"""

from typing import (
    Any,
    List,
    Dict,
    Set,
    Union,
    Optional,
)

from .built_in_rules import (
    BlockRule,
    CommentRule,
    EvalRule,
    InlineRule,
    IndentRule,
    ForInRule,
    JoinBlockRule,
    JoinEvalRule,
    FormatRule,
    StringifyRule,
)

from .rule import Rule
from .context import Context
from .exceptions import (
    InvalidTypeException,
    NoneValueException,
)


class Generator:
    """Manage rules and a generator for rules by dictionary"""

    STD_RULES = [
        BlockRule(),
        CommentRule(),
        EvalRule(),
        InlineRule(),
        IndentRule(),
        ForInRule(),
        JoinBlockRule(),
        JoinEvalRule(),
        FormatRule(),
        StringifyRule(),
    ]

    def __init__(
        self,
        gen_rules: List[Union[str, Dict[str, Any]]],
        parse_rules: Optional[List[Rule]] = None,
    ):
        """Initialization method for DictRule.

        Args:
            gen_rules (List[Union[str, Dict[str, Any]]]): List or dictionary of rules
            parse_rules (Optional[List[Rule]], optional):
                List of rule parsers bases on `Rule`.
                Defaults to `DictRule.STD_RULES`.
        """
        if parse_rules is None:
            parse_rules = Generator.STD_RULES

        self._cache_rules: Dict[str, Rule] = {}
        self._gen_rules = list(gen_rules)
        self._parse_rules: List[Rule] = []
        self.add_parse_rules(parse_rules)

    @property
    def parse_rules(self) -> List[Rule]:
        """Getter for `parse_rules` property"""

        return self._parse_rules

    @property
    def gen_rules(self) -> List[Union[str, Dict[str, Any]]]:
        """Getter for `gen_rules` property"""

        return self._gen_rules

    def add_parse_rule(
        self,
        rule: Rule,
    ):
        """Adds an extra rule to the parser.

        Args:
            rule (Rule): The rule subclass to add.
        """

        self._parse_rules.append(rule)

    def add_parse_rules(
        self,
        rules: List[Rule],
    ):
        """Adds a list of extra rules to parser.

        Args:
            rules (List[Rule]): List of rules, each being a subclass of Rule.
        """
        for rule in rules:
            self.add_parse_rule(rule)

    def generate(
        self,
        context: Optional[Context] = None,
    ) -> str:
        """Generate text based on `gen_rules`, `parse_rules` and `context`

        Args:
            context (Optional[Context], optional): The context to parse the rule.
                Defaults to None.
                @see `dicturle.Context`

        Returns:
            str: Generated text
        """

        def _rule_from_dict(rule_dict: Dict[str, Any]) -> Rule:
            rule = self._parse_rule_from_dict(rule_dict)
            if not rule:
                pass

            if rule is None:
                raise NoneValueException(f"Not found any rule in dict {rule_dict}")

            return rule

        def _internal_parse_rule(
            context: Optional[Context],
            rule: Any,
        ) -> str:
            if not rule:
                return ""

            if isinstance(rule, str):
                return rule

            if not isinstance(rule, Dict):
                raise InvalidTypeException(f"Rule {rule} must be a dict")

            found_rule = _rule_from_dict(rule)
            parsed = found_rule.parse(
                rule_dict=rule,
                context=context,
                rule_callback=_internal_parse_rule,
            )

            return parsed

        self._parse_rules.sort(
            key=lambda r: len(r.dr_non_optional_props),
            reverse=True,
        )

        output: List[str] = []
        for rule in self._gen_rules:
            parsed = _internal_parse_rule(
                context=context,
                rule=rule,
            )
            output.append(parsed)

        return "\n".join(output)

    def _parse_rule_from_dict(
        self,
        rule_dict: Dict[str, Any],
    ) -> Optional[Rule]:
        dict_keys = set(rule_dict.keys())
        key = self._key_from_keys(dict_keys)
        rule = self._cache_rules.get(key)
        if rule:
            return rule

        for rule in self._parse_rules:
            props: Set[str] = set()
            for prop in rule.dr_non_optional_props:
                prop_name, prop_value = prop(rule_dict)
                if not prop_value:
                    props = set()
                    break

                props.add(prop_name)

            if len(props) == 0:
                continue

            return rule

        return None

    def _key_from_keys(
        self,
        keys: Set[str],
    ) -> str:
        return ".".join(sorted(keys))



"""Join eval rule module"""

from typing import (
    Dict,
    Any,
    Callable,
    Iterable,
    Optional,
)

from .eval_rule import EvalRule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    InvalidTypeException,
)


class JoinEvalRule(EvalRule):
    """This rule joins generated texts from the value of `EvalRule` by a specified separator.

    Examples:
    ---------
    >>> context = build_football_teams_context([
    ...     "Real Madrid",
    ...     "Barcelona",
    ...     "Bayern Munchen",
    ...     "PSG",
    ...     "MC",
    ...     "MU",
    ...     "AC Milan",
    ... ])
    >>> dictrule.Generator({
    ...     "join": ", "
    ...     "eval": "football_teams"
    ... }).generate(context)
    Real Madrid, Barcelona, Bayern Munchen, PSG, MC, MU, AC Milan
    """

    @dr_property()
    def _join(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `join` attribute."""

    @dr_property()
    def _eval(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `eval` attribute."""

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parses the rule dictionary and returns the parsed block of text.

        Args:
            rule_dict (Dict[str, Any]): The rule dictionary to parse.
            rule_callback (Callable[[Optional[Context], Any], str]): A callback function
                to apply to each rule.
            context (Optional[Context], optional): The context to use during parsing.
                Defaults to None.

        Returns:
            str: The parsed block of text.
        """

        _, separator = self._join(rule_dict)
        if not isinstance(separator, str):
            raise InvalidTypeException("`join` must be a str")

        eval_rule = super().internal_parse(
            rule_dict=rule_dict,
            rule_callback=rule_callback,
            context=context,
        )

        if not isinstance(eval_rule, Iterable):
            raise InvalidTypeException("`join:eval:` must be a Iterable value")

        return separator.join(eval_rule)



"""Inline rule module"""

from typing import (
    Dict,
    Any,
    Callable,
    Optional,
)

from ..rule import Rule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    InvalidTypeException,
)


class InlineRule(Rule):
    """This rule builds generated text of a list of rules that are in a line.

    Examples:
    ---------
    >>> dictrule.Generator({
    ...    "inline": ["This", " is", " the", " text", " in", " a", " line"]
    ... }).generate()
    This is the text in a line
    """

    @dr_property()
    def _inline(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `inline` attribute."""

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parses the rule dictionary and returns the parsed string.

        Args:
            rule_dict (Dict[str, Any]): The rule dictionary to parse.
            rule_callback (Callable[[Optional[Context], Any], str]): A callback function
                to apply to each rule.
            context (Optional[Context], optional): The context to use during parsing.
                Defaults to None.

        Returns:
            str: The parsed string.
        """

        _, inline = self._inline(rule_dict)
        if not isinstance(inline, list):
            raise InvalidTypeException("`inline` must be a list")

        return "".join([str(rule_callback(context, rule)) for rule in inline])



r"""
    __          _ ____        _                     __         
   / /_  __  __(_) / /_      (_)___     _______  __/ /__  _____
  / __ \/ / / / / / __/_____/ / __ \   / ___/ / / / / _ \/ ___/
 / /_/ / /_/ / / / /_/_____/ / / / /  / /  / /_/ / /  __(__  ) 
/_.___/\__,_/_/_/\__/     /_/_/ /_/  /_/   \__,_/_/\___/____/  
                                                               
"""

from .block_rule import BlockRule
from .comment_rule import CommentRule
from .eval_rule import EvalRule
from .inline_rule import InlineRule
from .indent_rule import IndentRule
from .for_in_rule import ForInRule
from .join_block_rule import JoinBlockRule
from .join_eval_rule import JoinEvalRule
from .format_rule import FormatRule
from .stringify_rule import StringifyRule

__all__ = [
    "BlockRule",
    "CommentRule",
    "EvalRule",
    "InlineRule",
    "IndentRule",
    "ForInRule",
    "JoinBlockRule",
    "JoinEvalRule",
    "FormatRule",
    "StringifyRule",
]



"""Join block rule module"""

from typing import (
    Dict,
    Any,
    Callable,
    Optional,
)

from .block_rule import BlockRule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    InvalidTypeException,
)


class JoinBlockRule(BlockRule):
    """This rule joins generated texts for a block of rules by a specified separator.

    Examples:
    ---------
    >>> dictrule.Generator({
    ...     "join": "-"
    ...     "block": ["This", "is", "the", "snake", "line"]
    ... }).generate()
    This-is-the-snake-line
    """

    @dr_property()
    def _join(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `join` attribute."""

    @dr_property()
    def _block(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `block` attribute."""

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parses the rule dictionary and returns the parsed block of text.

        Args:
            rule_dict (Dict[str, Any]): The rule dictionary to parse.
            rule_callback (Callable[[Optional[Context], Any], str]): A callback function
                to apply to each rule.
            context (Optional[Context], optional): The context to use during parsing.
                Defaults to None.

        Returns:
            str: The parsed block of text.
        """

        _, join = self._join(rule_dict)
        if not isinstance(join, str):
            raise InvalidTypeException("`join` must be a str")

        block = super().internal_parse(
            rule_dict=rule_dict,
            rule_callback=rule_callback,
            separator=join,
            context=context,
        )

        return block



"""Comment rule module"""

from typing import (
    Dict,
    Any,
    Optional,
    Callable,
)

from enum import Enum
from ..dr_property import dr_property
from ..rule import Rule
from ..context import Context
from ..exceptions import (
    NoneValueException,
    InvalidTypeException,
    InvalidValueException,
)


class CommentRule(Rule):
    """This rule allows generating comment text based on single-line and multi-line rules.

    Building `Context` with `CommentRule.ContextCase` to define the comment style.

    Examples:
    ---------
    >>> context = build_singleline_context("#")
    >>> dictrule.Generator({
    ...     "comment": "This is a single-line comment"
    ... }).generate(context)
    "# This is a single-line comment"

    >>> context = build_multiline_context('\"""')
    >>> dictrule.Generator({
    ...     "style": "multiline",
    ...     "comment": [
    ...         "line1": "Line 1 of multi-line comment",
    ...         "line2": "Line 2 of multi-line comment",
    ...     ]
    ... }).generate(context)
    \"""
    Line 1 of multi-line comment
    Line 2 of multi-line comment
    \"""
    """

    CONTEXT_NAME = "comment"

    @dr_property()
    def _comment(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `comment` attribute."""

    @dr_property(optional=True)
    def _style(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `style` attribute."""

    class ContextCase(Context.Case):
        """`Context.Case` for `CommentRule`.

        This class defines different contexts for the CommentRule,
            including configurations for single-line and multi-line comments.

        Attributes:
            name (str): The name of the context, inherited from Context.Case.
            singleline (Optional[SinglelineComment]): Configuration for single-line comments.
            multiline (Optional[MultilineComment]): Configuration for multi-line comments.
        """

        class Comment:
            """Abstract class for `Commentrule.Comment`"""

            @staticmethod
            @property
            def style() -> "CommentRule.Style":
                """Comment style"""

                return None

        class SinglelineComment(Comment):
            """Singleline comment by prefix"""

            def __init__(
                self,
                prefix: str,
            ):
                """Initial method for `SinglelineComment`

                Args:
                    prefix (str): prefix for each line
                """

                self._prefix = prefix

            @staticmethod
            @property
            def style() -> "CommentRule.Style":
                return CommentRule.Style.SINGLELINE

            @property
            def prefix(self) -> str:
                """Getter for `prefix` property"""

                return self._prefix

        class MultilineComment(Comment):
            """Multiline comment by comment open, comment close, and prefix for each line"""

            def __init__(
                self,
                open_comment: str,
                close_comment: str,
                prefix: str,
            ):
                """Initial method for `MultilineComment`

                Args:
                    open_comment (str): open text of comment
                    close_comment (str): close text of comment
                    prefix (str): prefix for each line
                """

                self._prefix = prefix
                self._open_comment = open_comment
                self._close_comment = close_comment

            @staticmethod
            @property
            def style() -> "CommentRule.Style":
                return CommentRule.Style.MULTILINE

            @property
            def prefix(self) -> str:
                """Getter for `prefix` property"""

                return self._prefix

            @property
            def open_comment(self) -> str:
                """Getter for `open_comment` property"""

                return self._open_comment

            @property
            def close_comment(self) -> str:
                """Getter for `close_comment` property"""

                return self._close_comment

        @property
        def name(self) -> str:
            """Getter for `name` property"""

            return CommentRule.CONTEXT_NAME

        def __init__(
            self,
            singleline: Optional[SinglelineComment] = None,
            multiline: Optional[MultilineComment] = None,
        ):
            """Inital method for `ContextCase`

            Args:
                singleline (Optional[SinglelineComment], optional): singleline config.
                    Defaults to None.
                multiline (Optional[MultilineComment], optional): multiline config.
                    Defaults to None.
            """

            self._singleline = singleline
            self._multiline = multiline

        @property
        def singleline(self) -> SinglelineComment:
            """Getter for `singleline` property"""

            return self._singleline

        @property
        def multiline(self) -> MultilineComment:
            """Getter for `multiline` property"""

            return self._multiline

    class Style(Enum):
        """Enumeration defining comment styles"""

        SINGLELINE = "singleline"
        MULTILINE = "multiline"

        @staticmethod
        def from_str(string: str) -> Optional["CommentRule.Style"]:
            """Creates a comment style from string representation.

            Returns:
                Optional[CommentRule.Style]: The style of the comment.
            """

            try:
                return CommentRule.Style(string)
            except ValueError:
                pass

            return None

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parse the rule dictionary to generate comment text.

        Args:
            rule_dict (Dict[str, Any]): The dictionary of rules.
            rule_callback (Callable[[Optional[Context], Any], str]): Callback function
                for rules not handled by the current rule.
            context (Optional[Context], optional): The context for the rule. Defaults to None.

        Returns:
            str: The generated comment text.
        """

        if context is None:
            raise NoneValueException("Param `context` must not be None")

        context_case: CommentRule.ContextCase = context.get(CommentRule.CONTEXT_NAME)
        if not isinstance(context_case, CommentRule.ContextCase):
            raise InvalidTypeException(
                f"Invalid {type(context_case)} type for {CommentRule.CONTEXT_NAME} in context",
            )

        _, style_str = self._style(rule_dict)
        style = CommentRule.Style.from_str(style_str)
        if not style:
            style = CommentRule.Style.SINGLELINE

        _, comment = self._comment(rule_dict)
        if comment is None:
            raise NoneValueException("`comment` rule is invalid")

        comment_prefix = ""
        comment_open = ""
        comment_close = ""
        if style == CommentRule.Style.SINGLELINE:
            if context_case.singleline is None:
                raise NoneValueException(
                    f"Not found {CommentRule.ContextCase.SinglelineComment.__name__} \
                    from context"
                )

            prefix = context_case.singleline.prefix
            if not isinstance(prefix, str):
                raise InvalidTypeException(
                    f"Invalid prefix from context of {CommentRule.ContextCase.__name__} \
                    `{CommentRule.CONTEXT_NAME}`"
                )
            comment_prefix = prefix

        if style == CommentRule.Style.MULTILINE:
            if context_case.multiline is None:
                raise NoneValueException(
                    f"Not found {CommentRule.ContextCase.MultilineComment.__name__} \
                    from context"
                )

            comment_prefix = context_case.multiline.prefix
            comment_open = context_case.multiline.open_comment
            comment_close = context_case.multiline.close_comment

        rules_str = ""
        if isinstance(comment, dict):
            rules_str = rule_callback(context, comment)
            rules_str = rules_str.replace("\n", f"\n{comment_prefix}")
        elif isinstance(comment, str):
            rules_str = comment
            rules_str = rules_str.replace("\n", f"\n{comment_prefix}")
        elif isinstance(comment, list):
            rules_str = [rule_callback(context, rule) for rule in comment]
            rules_str = "\n".join(rules_str).replace("\n", f"\n{comment_prefix}")
        else:
            raise InvalidTypeException("`comment` rule is an invalid type")

        output = comment_prefix + rules_str

        if comment_open:
            output = comment_open + "\n" + output

        if comment_close:
            output += "\n" + comment_close

        return output



"""Module defines context for rule generation"""

from typing import (
    List,
    Dict,
    Optional,
)

from abc import (
    ABC,
)

from .exceptions import (
    NoneValueException,
    InvalidValueException,
)


class Context:
    """Context class contains information for rule generation"""

    class Case(ABC):
        """Base class for each rule, providing information for rule generation"""

        @property
        def name(self) -> str:
            """Getter for case name, matching the defined rule

            Returns:
                str: A rule name
            """

            return ""

    def __init__(
        self,
        cases: List[Case],
    ) -> None:
        """Initialization method of `Context` class

        Args:
            cases (List[Case]): List of `Context.Case` to build the case map
        """

        self._cases = list(cases)
        case_map: Dict[str, Context.Case] = {}
        for case in cases:
            if case.name in case_map:
                raise InvalidValueException(
                    f"Found existed {Context.Case.__name__}: {case.name}"
                )

            name = case.name
            if name is None:
                raise NoneValueException(f"Invalid case name `{name}`")

            case_map[case.name] = case

        self._case_map = case_map

    @property
    def cases(self) -> List[Case]:
        """Getter for `cases` property

        Returns:
            List[Case]: list of cases
        """
        return self._cases

    @property
    def case_map(self) -> Dict[str, Case]:
        """Getter for `case_map` property

        Returns:
            Dict[str, Case]: map of cases [name: str, case: Context.Case]
        """

        return self._case_map

    def get(self, name: str) -> Optional[Case]:
        """Gets the case by name

        Args:
            name (str): Name of the `Context.Case`

        Returns:
            Optional[Case]: `Context.Case` that matches to provided name,
                built from `Context` creation
        """

        return self._case_map.get(name)



"""Block rule module"""

from typing import (
    Dict,
    Any,
    Callable,
    Optional,
)

from ..rule import Rule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    InvalidTypeException,
)


class BlockRule(Rule):
    """This rule joins a list of rules separated by a new line.

    Examples:
    ---------
    >>> dictrule.Generator({
    ...   "block": [
    ...         "rule_1",
    ...         "rule_2",
    ...         "rule_3",
    ...     ]
    ... }).generate()
    rule_1
    rule_2
    rule_3
    """

    @dr_property()
    def _block(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `block` attribute."""

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parse the rule dictionary.

        Args:
            rule_dict (Dict[str, Any]): A dictionary of rules.
            rule_callback (Callable[[Optional[Context], Any], str]): Callback function
                for rules not handled by the current rule.
            context (Optional[Context], optional): Context for the rule. Defaults to None.

        Returns:
            str: The generated text.
        """

        return self.internal_parse(
            rule_dict=rule_dict,
            rule_callback=rule_callback,
            separator="\n",
            context=context,
        )

    def internal_parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        separator: str,
        context: Optional[Context] = None,
    ) -> str:
        """Internal parse method supporting custom separators instead of the new line as default.

        Args:
            rule_dict (Dict[str, Any]): Dictionary of rules to generate.
            rule_callback (Callable[[Optional[Context], Any], str]): Rule callback
                for rules not handled by the current rule.
            separator (str): Separator for each parsed rule from `rule_dict`.
            context (Optional[Context], optional): Context for the rule. Defaults to None.

        Returns:
            str: The generated text.
        """

        _, block = self._block(rule_dict)
        if not isinstance(block, list):
            raise InvalidTypeException("`block` value must be a list")

        return separator.join([str(rule_callback(context, rule)) for rule in block])



r"""
       ___      __             __   
  ____/ (_)____/ /________  __/ /__ 
 / __  / / ___/ __/ ___/ / / / / _ \
/ /_/ / / /__/ /_/ /  / /_/ / /  __/
\__,_/_/\___/\__/_/   \__,_/_/\___/ 

"""

from pathlib import Path
from setuptools import setup

this_directory = Path(__file__).parent

about = {}
with open(
    file=this_directory / "src" / "dictrule" / "__version__.py",
    mode="r",
    encoding="utf-8",
) as file:
    exec(file.read(), about)

long_description = (this_directory / "README.md").read_text()

if __name__ == "__main__":
    setup(
        name=about["__title__"],
        version=about["__version__"],
        description=about["__description__"],
        long_description=long_description,
        long_description_content_type="text/markdown",
        author=about["__author__"],
        author_email=about["__author_email__"],
        url=about["__url__"],
        packages=["dictrule", "dictrule/built_in_rules"],
        package_data={"": ["LICENSE", "NOTICE"]},
        package_dir={"": "src"},
        include_package_data=True,
        python_requires=">=3.7",
        license=about["__license__"],
        zip_safe=False,
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Text Processing :: Markup",
        ],
        project_urls={
            "Documentation": "https://github.com/elhoangvu/dictrule",
            "Source": "https://github.com/elhoangvu/dictrule",
        },
    )



r"""
       ___      __             __   
  ____/ (_)____/ /________  __/ /__ 
 / __  / / ___/ __/ ___/ / / / / _ \
/ /_/ / / /__/ /_/ /  / /_/ / /  __/
\__,_/_/\___/\__/_/   \__,_/_/\___/ 

"""

from .generator import Generator
from .rule import Rule
from .context import Context
from .__version__ import (
    __title__,
    __description__,
    __url__,
    __version__,
    __author__,
    __author_email__,
    __license__,
    __copyright__,
)

from .exceptions import (
    NoneValueException,
    InvalidTypeException,
    InvalidValueException,
)

from .built_in_rules import (
    BlockRule,
    CommentRule,
    EvalRule,
    InlineRule,
    IndentRule,
    ForInRule,
    JoinBlockRule,
    JoinEvalRule,
    FormatRule,
)

__all__ = [
    "Generator",
    "Rule",
    "Context",
    "NoneValueException",
    "InvalidTypeException",
    "InvalidValueException",
    "BlockRule",
    "CommentRule",
    "EvalRule",
    "InlineRule",
    "IndentRule",
    "ForInRule",
    "JoinBlockRule",
    "JoinEvalRule",
    "FormatRule",
    "__title__",
    "__description__",
    "__url__",
    "__version__",
    "__author__",
    "__author_email__",
    "__license__",
    "__copyright__",
]



"""Stringify rule module"""

from typing import (
    Dict,
    Any,
    Callable,
    Optional,
)

from ..rule import Rule
from ..dr_property import dr_property
from ..context import Context


class StringifyRule(Rule):
    """This rule wraps content by quotes.

    Examples:
    ---------
    >>> dictrule.Generator({
    ...    "stringify": "This is the 1st text"
    ...    "stringify": "This is the 2nd text"
    ... }).generate()
    "This is the 1st text"
    "This is the 2nd text"
    """

    QUOTE = '"'

    @dr_property()
    def _stringify(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `stringify` attribute."""

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Stringify a text by enclosing it with specified quotes.

        Args:
            rule_dict (Dict[str, Any]): The rule dictionary to parse.
            rule_callback (Callable[[Optional[Context], Any], str]): A callback function
                to apply to each rule.
            context (Optional[Context], optional): The context to use during parsing.
                Defaults to None.

        Returns:
            str: The parsed string.
        """

        _, stringify = self._stringify(rule_dict)
        value = stringify
        if not isinstance(value, str):
            value = rule_callback(context, value)

        if value is None:
            value = ""

        return StringifyRule.QUOTE + str(value) + StringifyRule.QUOTE



"""Eval rule module"""

from typing import (
    Dict,
    List,
    Any,
    Callable,
    Optional,
)

from abc import ABC, abstractmethod
from functools import lru_cache
from ..rule import Rule
from ..dr_property import dr_property
from ..context import Context
from ..exceptions import (
    NoneValueException,
    InvalidTypeException,
)


class EvalRule(Rule):
    """This rule evaluates the value of provied variable or command.

    Examples:
    ---------
    >>> git_author = "Zooxy Le"
    >>> dictrule.Generator({
    ...     "eval": "git_author"
    ... }).generate()
    Zooxy Le

    >>> project_id = "123456789"
    >>> dictrule.Generator({
    ...     "eval": "project_id"
    ... }).generate()
    123456789
    """

    CONTEXT_NAME = "eval"

    @dr_property()
    def _eval(self, props: Dict[str, Any]) -> Any:
        """Property method for retrieving the `eval` attribute."""

    class Evaluable(ABC):
        """Abstract class for the evaluator."""

        @property
        def name(self) -> str:
            """Name or command for evaluator"""

            return EvalRule.CONTEXT_NAME

        @property
        def prefix_matching(self) -> bool:
            """`prefix_matching` property defines evaluator should use prefix for matching name"""
            return False

        @abstractmethod
        def run(
            self,
            cmd: str,
        ) -> Any:
            """Run `cmd`

            Args:
                cmd (str): command or variable name to eval the vale

            Returns:
                Any: value of `cmd`
            """

            return None

    class KeyValueEvaluator(Evaluable):
        """Evaluator using key-value pairs."""

        def __init__(
            self,
            key: str,
            value: Any,
        ):
            """Initial method for `KeyValueEvaluator`

            Args:
                key (str): key as name of `Evaluable`
                value (Any): value of eval name
            """

            self._key = key
            self._value = value

        @property
        def name(self) -> str:
            return self._key

        def run(self, cmd: str) -> Any:
            return self._value

    class ContextCase(Context.Case):
        """`Context.Case` for `EvalRule`.

        This class manages contexts for the `EvalRule`,
        including a list of evaluators and fallbacks.

        Attributes:
            name (str): The name of the context, inherited from `Context.Case`.
            evaluator_list (List["EvalRule.Evaluable"]): A list of evaluators.
            fallback (Optional["EvalRule.Evaluable"]): The fallback evaluator for other rules.

        Args:
            evaluators (List["EvalRule.Evaluable"]): The list of evaluators.
            fallback (Optional["EvalRule.Evaluable"], optional): The fallback evaluator
                for other rules. Defaults to None.
        """

        @property
        def name(self) -> str:
            """Getter for `name` property"""

            return EvalRule.CONTEXT_NAME

        @property
        def evaluator_list(self) -> List["EvalRule.Evaluable"]:
            """Getter for `evaluator_list` property"""

            return self._evaluator_list

        @property
        def fallback(self) -> Optional["EvalRule.Evaluable"]:
            """Getter for `fallback` property"""

            return self._fallback

        def __init__(
            self,
            evaluators: List["EvalRule.Evaluable"],
            fallback: Optional["EvalRule.Evaluable"] = None,
        ):
            """Initialization method for `EvalRule.ContextCase`.

            Args:
                evaluators (List["EvalRule.Evaluable"]): List of evaluators.
                fallback (Optional["EvalRule.Evaluable"], optional): Fallback for other rules.
                    Defaults to None.
            """

            self._evaluator_list = list(evaluators)
            self._fallback = fallback
            nonprefix_evaluators: Dict[str, EvalRule.Evaluable] = {}
            prefix_evaluators: Dict[str, EvalRule.Evaluable] = {}

            for evaluator in evaluators:
                if evaluator.prefix_matching:
                    prefix_evaluators[evaluator.name] = evaluator
                else:
                    nonprefix_evaluators[evaluator.name] = evaluator

            self._evaluators = nonprefix_evaluators
            self._prefix_evaluators = prefix_evaluators

        def eval(
            self,
            eval_name: str,
        ) -> Optional[Any]:
            """Evaluates value with name.

            Args:
                eval_name (str): The name for evaluating.

            Returns:
                Optional[Any]: Evaluated value.
            """

            eval_rule = self._evaluators.get(eval_name)
            if not eval_rule:
                eval_rule = self._find_eval_by_prefix(
                    eval_name=eval_name,
                )

            if not eval_rule:
                if self._fallback:
                    return self._fallback.run(eval_name)
                return None

            return eval_rule.run(eval_name)

        @lru_cache(maxsize=1024)
        def _find_eval_by_prefix(
            self,
            eval_name: str,
        ) -> Optional["EvalRule.Evaluable"]:
            """Finds an evaluator by prefix.

            Args:
                eval_name (str): The name for evaluation.

            Returns:
                Optional[EvalRule.Evaluable]: The evaluator found by prefix.
            """

            for name, evaluator in self._prefix_evaluators.items():
                if eval_name.startswith(name):
                    return evaluator

            return None

    def parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> str:
        """Parses the rule dictionary for EvalRule.

        Args:
            rule_dict (Dict[str, Any]): Dictionary of rules containing EvalRule.
            rule_callback (Callable[[Optional[Context], Any], str]): Fallback for other rules.
            context (Optional[Context], optional): EvalRule's context. Defaults to None.

        Returns:
            str: Parsed value for `rule_dict`.
        """

        parsed = self.internal_parse(
            rule_dict=rule_dict,
            rule_callback=rule_callback,
            context=context,
        )

        if parsed is None:
            raise NoneValueException(
                f"EvalRule with dict `{rule_dict}` reacts None value"
            )

        return parsed

    def internal_parse(
        self,
        rule_dict: Dict[str, Any],
        rule_callback: Callable[[Optional[Context], Any], str],
        context: Optional[Context] = None,
    ) -> Any:
        """Internal method for parsing rule dictionary for EvalRule.

        Args:
            rule_dict (Dict[str, Any]): Dictionary of rules containing EvalRule.
            rule_callback (Callable[[Optional[Context], Any], str]): Fallback for other rules.
            context (Optional[Context], optional): EvalRule's context. Defaults to None.

        Returns:
            Any: Parsed value for `rule_dict`.
        """

        _ = rule_callback
        if context is None:
            raise NoneValueException("param `context` must not be None")

        context_case: EvalRule.ContextCase = context.get(EvalRule.CONTEXT_NAME)

        if not isinstance(context_case, EvalRule.ContextCase):
            raise InvalidTypeException(
                f"Invalid {type(context_case)} type for {EvalRule.CONTEXT_NAME} in context"
            )

        _, eval_rule = self._eval(rule_dict)
        if not isinstance(eval_rule, str):
            raise InvalidTypeException("`eval` must be a str")

        value = context_case.eval(eval_rule)
        return value
